"""
Validation module:

Contains routines that validate inputs for both the masked_norm and
affine_masked_norm procedures

The routines presented in this module make no effort in validating the input
objects types. Code is type annotated, which doesn't replace proper
validation at runtime.
"""

from __future__ import annotations

from torch import Tensor, Size
from torch import float32, float64, float16, bfloat16
from torch import float8_e4m3fn, float8_e5m2
from torch import bool as bool_dtype


# PyTorch doesn't offer abstract type hinting for xfloat* dtypes
# you can work around this by verifying membership on a set of xfloat* dtypes
float_dtype = {
    float64,
    float32,
    float16,
    bfloat16,
    float8_e4m3fn,
    float8_e5m2
}


def validate_masked_norm(inpt: Tensor, mask: Tensor | None) -> Tensor:
    """
    Validates the input of masked_norm procedure
    Outputs are reshaped according to mask
    """

    if not inpt.dtype in float_dtype:
        raise ValueError("'inpt' must be a tensor of xfloat* dtype")

    if mask is None:

        return inpt

    if not mask.dtype == bool_dtype:
        # mask tensor must be necessarily of bool dtype
        raise ValueError("'mask' must be a tensor of bool dtype")

    if len(inpt.shape) < len(mask.shape):
        raise ValueError(
            "'mask' must have a smaller number of axes than 'inpt'"
        )

    # mask shape must partially match input shape
    for axis, ax_dim in enumerate(zip(inpt.shape, mask.shape)):
        inpt_ax_dim, mask_ax_dim = ax_dim
        if inpt_ax_dim != mask_ax_dim:
            raise ValueError(
                f"dim mismatch between 'inpt' and 'mask' on axis {axis}"
        )

    return inpt[mask]


def validate_affine_masked_norm(
    inpt: Tensor,
    mask: Tensor | None,
    weight: Tensor,
    bias: Tensor | None
) -> tuple[Tensor, Tensor, Tensor | None]:
    """
    Validates the input of the affine_masked_norm procedure
    Outputs are reshaped according to mask
    """

    shape: Size

    out = validate_masked_norm(inpt, mask)

    # weight and bias must have proper dtype
    if not weight.dtype in float_dtype:
        raise ValueError("'weight' must be a tensor of xfloat* dtype")

    if not bias is None and not bias.dtype in float_dtype:
        raise ValueError("'bias' must be a tensor of xfloat* dtype")


    if mask is None:
        # if no mask is passed, weight and bias tensor must be flat
        shape = out.shape[0: 1]

        if weight.shape != shape:
            raise ValueError(
                f"'weight' must be a flat tensor with shape {shape}"
            )

        if not bias is None and bias.shape != shape:
            raise ValueError(
                f"'bias' must be a flat tensor with shape {shape}"
            )

        return out, weight, bias

    # if a mask is passed, weight and bias must have the same shape as mask
    shape = mask.shape

    if weight.shape != shape:
        raise ValueError(
            "'weight' must have the same shape as 'mask'"
        )

    if not bias is None and bias.shape != shape:
        raise ValueError(
            "'bias' must have the same shape as 'mask'"
        )

    if bias is None:
        return out, weight[mask], bias

    return out, weight[mask], bias[mask]



"""
Entry-point module:

You can import any of the relevant objects by simply:
>>> from masked_norm import LazyAffineMaskedNorm
"""

from .functional import masked_norm
from .nn import MaskedNorm
from .nn import LazyMaskedNorm

from .functional import affine_masked_norm
from .nn import LazyAffineMaskedNorm



"""
Utilities module:

Contains simple utilities routine
"""

from torch import Tensor


def get_factory_key(key: str) -> bool:
    """
    Filtering procedure for selecting factory kwargs
    """

    if key == "device":
        return True
    if key == "dtype":
        return True
    return False


def batchwise_add(flat: Tensor, x: Tensor) -> Tensor:
    """
    """

    if flat.shape != x.shape[0: 1]:
        raise ValueError("dim mismatch on the first axis")

    for _ in x.shape[1:]:
        flat = flat[None, ...]

    flat = flat.transpose(0, -1)

    return flat + x


def batchwise_mul(flat: Tensor, x: Tensor) -> Tensor:
    """
    """

    if flat.shape != x.shape[0: 1]:

        print(flat.shape, x.shape)
        raise ValueError("dim mismatch on the first axis")

    for _ in x.shape[1:]:
        flat = flat[None, ...]

    flat = flat.transpose(0, -1)

    return flat * x



"""
Torch nns module:

Contains implementation of PyTorch Modules that perform masked normalization
"""

from __future__ import annotations

from typing import Any

from torch import Tensor
from torch.nn import Module
from torch.nn.modules.lazy import LazyModuleMixin
from torch.nn.parameter import UninitializedParameter
from torch.nn.init import ones_, zeros_

from .functional import masked_norm
from .functional import affine_masked_norm
from .util import get_factory_key


class MaskedNorm(Module):
    """
    MaskedNorm PyTorch class definition

    Plain masked normalization is stateless, but many DL devs prefer their use
    over the functional form
    """

    def __init__(
        self: MaskedNorm,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        """

        super().__init__(*args, **kwargs)

    def forward(
        self: MaskedNorm,
        inpt: Tensor,
        mask: Tensor | None = None,
    ) -> Tensor:
        """
        Forward method
        """

        return masked_norm(inpt, mask)


class LazyMaskedNorm(MaskedNorm, LazyModuleMixin):
    """
    LazyMaskedNorm PyTorch class defintion

    Stateful implementation of MaskedNorm is crazy, but a lazy variation of
    a MaskedNorm stateful implementation is absolutely ludicrous
    Although you can use it, it simply serves as inheritence reference for
    LazyAffineMaskedNorm - which is stateful
    """

    def __init__(
        self: LazyMaskedNorm,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        """

        super().__init__(*args, **kwargs)
        super(MaskedNorm, self).__init__()

    def initialize_parameters(
        self: LazyMaskedNorm,
        inpt: Tensor,
        mask: Tensor | None = None,
    ) -> None:
        """
        Lazy-specific method for initializing parameters according to the
        input
        """

        pass


class LazyAffineMaskedNorm(LazyMaskedNorm):
    """
    LazyAffineMaskedNorm PyTorch class defintion

    As a lazy module, the weight and bias parameters are initialized
    dynamically as the forward method is called
    """

    # affine transformation parameters
    weight: UninitializedParameter
    bias: UninitializedParameter

    def __init__(
        self: LazyAffineMaskedNorm,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        """

        super().__init__(*args, **kwargs)

        # the module's weight and bias parameters must be initialized with the
        # correct "device" and "dtype" that are passed onto the module
        factory_kwargs = dict(filter(get_factory_key, kwargs))

        self.weight = UninitializedParameter(**factory_kwargs)
        self.bias = UninitializedParameter(**factory_kwargs)

    def initialize_parameters(
        self: LazyAffineMaskedNorm,
        inpt: Tensor,
        mask: Tensor | None = None,
    ) -> None:
        """
        Lazy-specific method for initializing parameters according to the
        input
        """

        if self.has_uninitialized_params():

            # allocate memory to the parameters accordingly
            if mask is None:
                shape = inpt.shape[0: 1]
                self.weight.materialize(shape)
                self.bias.materialize(shape)
            else:
                self.weight.materialize(mask.shape)
                self.bias.materialize(mask.shape)

            # fill allocated parameters with initial values
            ones_(self.weight)
            zeros_(self.bias)


    def forward(
        self: LazyMaskedNorm,
        inpt: Tensor,
        mask: Tensor | None = None
    ) -> Tensor:
        """
        Forward method
        """

        self.initialize_parameters(inpt, mask)

        return affine_masked_norm(inpt, mask, self.weight, self.bias)



"""
Functional module:

Contains the functional implementation of masked normalization
"""

from __future__ import annotations

from torch import Tensor

from .validation import validate_masked_norm
from .validation import validate_affine_masked_norm
from .util import batchwise_mul, batchwise_add


def masked_norm(
    inpt: Tensor,
    mask: Tensor | None = None,
) -> Tensor:
    """
    Masked normalization procedure

    Normalizes only elements of the input specified by the mask; if no mask is
    passed to the routine, normalization is performed across the first axis
    If either by chance or design a collection of samples yields null
    variance, the normalization over such collection is ignored, and the
    values are passed along unaltered
    """

    out = validate_masked_norm(inpt, mask)

    clone = inpt.clone()

    dim = out.dim()
    reduce = tuple(range(1, dim))

    mean = out.mean(dim=reduce, keepdim=True)
    var = out.var(dim=reduce, keepdim=True)

    var_mask = (var != 0.0).flatten()

    out = (out[var_mask] - mean[var_mask]) / var[var_mask].sqrt()

    if mask is None:
        clone[var_mask] = out
    else:
        tmp = clone[mask].clone()
        tmp[var_mask] = out
        clone[mask] = tmp

    return clone


def affine_masked_norm(
    inpt: Tensor,
    mask: Tensor | None,
    weight: Tensor,
    bias: Tensor | None,
) -> Tensor:
    """
    Affine masked normalization procedure

    Normalizes only elements of the input specified by the mask; if no mask is
    passed to the routine, normalization is performed across the first axis
    If either by chance or design a collection of samples yields null
    variance, the normalization over such collection is ignored, and the
    values are passed along unaltered

    After normalization an affine transformation is performed along the
    normalized dimensions
    """

    out, weight, bias = validate_affine_masked_norm(inpt, mask, weight, bias)

    clone = inpt.clone()

    dim = out.dim()
    reduce = tuple(range(1, dim))

    mean = out.mean(dim=reduce, keepdim=True)
    var = out.var(dim=reduce, keepdim=True)

    var_mask = (var != 0.0).flatten()

    out = (out[var_mask] - mean[var_mask]) / var[var_mask].sqrt()

    out = batchwise_mul(weight[var_mask], out)

    if not bias is None:
        out = batchwise_add(bias[var_mask], out)

    if mask is None:
        clone[var_mask] = out
    else:
        tmp = clone[mask].clone()
        tmp[var_mask] = out
        clone[mask] = tmp

    return clone



"""
Validation module testing
"""

from __future__ import annotations

from torch import Tensor, tensor

from .tensor_test_case import TensorTestCase

from masked_norm.validation import validate_masked_norm
from masked_norm.validation import validate_affine_masked_norm


class TestValidateMaskedNorm(TensorTestCase):
    """
    'validate_masked_norm' testing unit
    """

    def setUp(self: TestValidateMaskedNorm) -> None:
        """
        """

        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_mask_0(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_0 = None

        out = validate_masked_norm(self.inpt, mask_0)

        self.assertEqTensor(out, self.inpt)

    def test_mask_1(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        out_1 = tensor(
            [
                [0.00, 1.00, 2.00]
            ]
        )

        out = validate_masked_norm(self.inpt, mask_1)

        self.assertEqTensor(out, out_1)

    def test_mask_2(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_2 = tensor(
            [
                [True, True, False],
                [True, False, False]
            ]
        )

        out_2 = tensor(
            [
                0.00,
                1.00,
                3.00
            ]
        )

        out = validate_masked_norm(self.inpt, mask_2)

        self.assertEqTensor(out, out_2)

    def test_mask_3(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_3 = tensor(
            [
                [
                    [True]
                ]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'mask' must have a smaller number of axes than 'inpt'"
        ):
            validate_masked_norm(self.inpt, mask_3)

    def test_mask_4(self: TestValidateMaskedNorm) -> None:
        """
        """

        mask_4 = tensor(
            [
                [True],
                [False]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "dim mismatch between 'inpt' and 'mask'"
        ):

            validate_masked_norm(self.inpt, mask_4)


class TestValidateAffineMaskedNorm(TensorTestCase):
    """
    'validate_affine_masked_norm' testing unit
    """

    def setUp(self: TestValidateAffineMaskedNorm) -> None:
        """
        """


        self.inpt = tensor(
            [
                [0.00, 1.00, 2.00],
                [3.00, 4.00, 5.00]
            ]
        )

    def test_mask_0_weight_1_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_0 = None

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_0,
            weight_1,
            bias_0
        )

        self.assertEqTensor(weight, weight_1)
        self.assertEqTensor(bias, bias_0)

    def test_mask_0_weight_1_bias_1(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_1 = tensor(
            [0.0, 0.0]
        )

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_0,
            weight_1,
            bias_1
        )

        self.assertEqTensor(weight, weight_1)
        self.assertEqTensor(bias, bias_1)

    def test_mask_0_weight_2_bias_0(
            self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_2 = tensor(
            [
                [1.0, 1.0, 1.0]
            ]
        )

        bias_0 = None

        with self.assertRaisesRegex(
            ValueError,
            "'weight' must be a flat tensor"
        ):
            validate_affine_masked_norm(self.inpt, mask_0, weight_2, bias_0)

    def test_mask_0_weigth_1_bias_2(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_0 = None

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_2 = tensor(
            [
                [0.0, 0.0, 0.0]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'bias' must be a flat tensor"
        ):
            validate_affine_masked_norm(self.inpt, mask_0, weight_1, bias_2)

    def test_mask_1_weight_1_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_0 = None

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_1,
            weight_1,
            bias_0
        )

        self.assertEqTensor(weight, weight_1)
        self.assertEqTensor(bias, bias_0)

    def test_mask_1_weight_1_bias_1(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_1 = tensor(
            [0.0, 0.0]
        )

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_1,
            weight_1,
            bias_1
        )

        self.assertEqTensor(weight, weight_1)
        self.assertEqTensor(bias, bias_1)

    def test_mask_1_weight_2_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_2 = tensor(
            [
                [1.0, 1.0, 1.0]
            ]
        )

        bias_0 = None

        with self.assertRaisesRegex(
            ValueError,
            "'weight' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_1, weight_2, bias_0)

    def test_mask_1_weight_1_bias_2(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_1 = tensor(
            [True, False]
        )

        weight_1 = tensor(
            [1.0, 1.0]
        )

        bias_2 = tensor(
            [
                [0.0, 0.0, 0.0]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'bias' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_1, weight_1, bias_2)

    def test_mask_2_weight_3_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_2 = tensor(
            [
                [True, True, False],
                [True, False, False]
            ]
        )

        weight_3 = tensor(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0]
            ]
        )

        bias_0 = None

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_2,
            weight_3,
            bias_0
        )

        self.assertEqTensor(weight, weight_3)
        self.assertEqTensor(bias, bias_0)

    def test_mask_2_weight_3_bias_3(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_2 = tensor(
            [
                [True, True, False],
                [True, False, False]
            ]
        )

        weight_3 = tensor(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0]
            ]
        )

        bias_3 = tensor(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0]
            ]
        )

        out, weight, bias = validate_affine_masked_norm(
            self.inpt,
            mask_2,
            weight_3,
            bias_3
        )

        self.assertEqTensor(weight, weight_3)
        self.assertEqTensor(bias, bias_3)

    def test_mask_2_weight_4_bias_0(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_2 = tensor(
            [
                [True, True, False],
                [True, False, False]
            ]
        )

        weight_4 = tensor(
            [
                [1.0, 1.0],
                [1.0, 1.0]
            ]
        )

        bias_0 = None

        with self.assertRaisesRegex(
            ValueError,
            "'weight' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_2, weight_4, bias_0)

    def test_mask_2_weight_3_bias_4(
        self: TestValidateAffineMaskedNorm
    ) -> None:
        """
        """

        mask_2 = tensor(
            [
                [True, True, False],
                [True, False, False]
            ]
        )

        weight_3 = tensor(
            [
                [1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0]
            ]
        )

        bias_4 = tensor(
            [
                [0.0, 0.0],
                [0.0, 0.0]
            ]
        )

        with self.assertRaisesRegex(
            ValueError,
            "'bias' must have the same shape as 'mask'"
        ):
            validate_affine_masked_norm(self.inpt, mask_2, weight_3, bias_4)
